(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_auth.ml 187 2005-10-14 10:54:15Z gerd $ *)

open Umlmon_types
open Umlmonctrl_aux
open Umlmon_common
open Cryptokit

module Conn = struct
  type t = Rpc_server.connection_id
  let compare : t -> t -> int = Pervasives.compare
end


module ConnMap = Map.Make(Conn)


let auth_conns = (ref ConnMap.empty : 
		    [ `Unauthorised | `Failed | `Authorised | `Challenge of string ] ConnMap.t ref )


let seed = lazy(
  Random.string Random.secure_rng 64
);;


let ensure_authenticated sess =
  try
    let info = ConnMap.find (Rpc_server.get_connection_id sess) !auth_conns in
    match info with
      | `Unauthorised ->
	  raise (Auth_failed (Rpc_server.get_peer_name sess))
      | `Failed ->
	  raise (Auth_failed (Rpc_server.get_peer_name sess))
      | `Challenge _ ->
	  raise (Auth_failed (Rpc_server.get_peer_name sess))
      | `Authorised ->
	  ()
  with
      Not_found ->
	( match Rpc_server.get_socket_name sess with
	    | Unix.ADDR_UNIX _ ->
		()
	    | Unix.ADDR_INET _ ->
		raise (Auth_failed (Rpc_server.get_peer_name sess))
	)

let forget_connection conn_id =
  auth_conns := ConnMap.remove conn_id !auth_conns

let known_as_failed conn_id =
  try
    ConnMap.find conn_id !auth_conns = `Failed
  with
      Not_found -> false

let proc_auth_get_challenge gs =
  protect
    (fun sess () reply ->
       let s = Random.string (Random.pseudo_rng (Lazy.force seed)) 64 in
       let conn_id = Rpc_server.get_connection_id sess in
       ( try
	   let info = ConnMap.find conn_id !auth_conns in
	   match info with
	     | `Unauthorised ->
		 raise Not_found
	     | `Failed ->
		 reply s   (* but don't record [s]! *)
	     | `Challenge _ ->
		 reply s   (* but don't record [s]! *)
	     | `Authorised ->
		 raise Not_found
	 with
	     Not_found ->
	       auth_conns := ConnMap.add conn_id (`Challenge s) !auth_conns;
	       reply s
       )
    )

let proc_auth_pass_credentials ues gs =
  protect
    (fun session (algoname,responses) reply ->
       let globalspec = Lazy.force gs.umlglobal in
       let spec = Lazy.force gs.umlspec in
       let conn_id = Rpc_server.get_connection_id session in
       let responses = Array.to_list responses in
       try
	 if algoname <> "HMAC-SHA1" then
	   failwith "Unknown HMAC algorithm";
	 let challenge =
	   try
	     let info = ConnMap.find conn_id !auth_conns in
	     match info with
	       | `Challenge s -> s
	       | `Unauthorised
	       | `Authorised
	       | `Failed -> raise Not_found
	   with
	       Not_found -> 
		 failwith "No challenge known" in
	 let expected_response1 =
	   match spec.umlvmuserpw with
	       None -> None
	     | Some pw -> 
		 let mac = MAC.hmac_sha1 challenge in
		 let h = hash_string mac pw in
		 (Some h) in
	 let expected_response2 =
	   match globalspec.umlmasterpw with
	     | None -> None
	     | Some pw ->
		 let mac = MAC.hmac_sha1 challenge in
		 Some(hash_string mac pw) in
	 let expected_response3 =
	   match spec.umlvmuserpw with
	       None -> None
	     | Some pw -> 
		 let mac = MAC.hmac_sha1 challenge in
		 let h = hash_string mac pw in
		 (Some ("userpw:" ^ h)) in
	 
	 if List.exists(fun response ->
			  Some response = expected_response1 || 
		           Some response = expected_response2 ||
			  Some response = expected_response3 ) responses 
	 then (
	   auth_conns := ConnMap.add conn_id `Authorised !auth_conns;
	   reply true
	 )
	 else
	   failwith "Authentication error"
       with
	   Failure msg ->
	     auth_conns := ConnMap.add conn_id `Failed !auth_conns;
	     let g = Unixqueue.new_group ues in
	     Unixqueue.once ues g 3.0 (fun () -> reply false);
	     monitor_log "umlmon: Failed authentication";
    )


let proc_auth_reset ues gs =
  protect
    (fun session () reply ->
       ensure_authenticated session;
       let conn_id = Rpc_server.get_connection_id session in
       auth_conns := ConnMap.add conn_id `Unauthorised !auth_conns;
       reply ()
    )


let proc_auth_change_password ~write_vm_config gs =
  protect
    (fun session enc_new_pw reply ->
       ensure_authenticated session;
       let globalspec = Lazy.force gs.umlglobal in
       let spec = Lazy.force gs.umlspec in
       try
	 if String.length enc_new_pw <> 16 then
	   failwith "Bad argument length";
	 let new_pw =
	   match spec.umlvmuserpw with
	       None ->
		 failwith "User does not have password, cannot change it"
	     | Some pw -> 
		 let trans = Cipher.aes ~mode:Cipher.CBC pw Cipher.Decrypt in
		 transform_string trans enc_new_pw
	 in
	 let spec' = { spec with umlvmuserpw = Some new_pw } in
	 write_vm_config spec.umlid spec';
	 gs.umlspec <- lazy spec';
	 reply `code_ok
       with
	   Failure msg ->
	     reply (`code_error msg);
	     monitor_log "umlmon: Failed to change password";
    )
